using NUnit.Framework;
using System;

namespace Lydos.Matlab.Tests
{
    [TestFixture]
    public class MatrixTests
    {
        #region Fields
        Matrix m;
        #endregion

        #region Set up and tear down
        [TestFixtureSetUp]
        public void Init()
        {
            m = new Matrix(2, 3);

            m[0, 0] = 0;
            m[0, 1] = 1;
            m[0, 2] = 2;
            m[1, 0] = 3;
            m[1, 1] = 4;
            m[1, 2] = 5;
        }

        [TestFixtureTearDown]
        public void Dispose()
        {
            m.Dispose();
        }
        #endregion

        #region Unit tests
        [Test]
        public void Acessors()
        {
            for (int i = 0; i < m.Rows; ++i)
            {
                for (int j = 0; j < m.Cols; ++j)
                {
                    Assert.AreEqual(m[i, j], m[j * m.Rows + i]);
                }
            }
        }

        [Test]
        public void CreateEmptyInit()
        {
            Matrix n = new Matrix(2, 2);

            Assert.AreEqual(0, n[0, 0]);
            Assert.AreEqual(0, n[0, 1]);
            Assert.AreEqual(0, n[1, 0]);
            Assert.AreEqual(0, n[1, 1]);
            Assert.IsFalse(n.IsScalar);
            Assert.IsFalse(n.IsVector);
            Assert.IsFalse(n.IsColVector);
            Assert.IsFalse(n.IsRowVector);
            Assert.IsTrue(n.IsMatrix);

            n.Dispose();
        }

        [Test]
        public void FromArrayColMajor()
        {
            double[] d = new double[6] {0, 3, 1, 4, 2, 5};
            Matrix n = Matrix.FromArray(d, 3, false);

            Assert.AreEqual(n.Rows, 2);
            Assert.AreEqual(n.Cols, 3);
            Assert.AreEqual(d[0], n[0, 0]);
            Assert.AreEqual(d[1], n[1, 0]);
            Assert.AreEqual(d[2], n[0, 1]);
            Assert.AreEqual(d[3], n[1, 1]);
            Assert.AreEqual(d[4], n[0, 2]);
            Assert.AreEqual(d[5], n[1, 2]);
            Assert.IsFalse(n.IsScalar);
            Assert.IsFalse(n.IsVector);
            Assert.IsFalse(n.IsColVector);
            Assert.IsFalse(n.IsRowVector);
            Assert.IsTrue(n.IsMatrix);

            n.Dispose();
        }

        [Test]
        public void FromArrayRowMajor()
        {
            double[] d = new double[6] {0, 1, 2, 3, 4, 5};
            Matrix n = Matrix.FromArray(d, 3);

            Assert.AreEqual(n.Rows, 2);
            Assert.AreEqual(n.Cols, 3);
            Assert.AreEqual(d[0], n[0, 0]);
            Assert.AreEqual(d[1], n[0, 1]);
            Assert.AreEqual(d[2], n[0, 2]);
            Assert.AreEqual(d[3], n[1, 0]);
            Assert.AreEqual(d[4], n[1, 1]);
            Assert.AreEqual(d[5], n[1, 2]);
            Assert.IsFalse(n.IsScalar);
            Assert.IsFalse(n.IsVector);
            Assert.IsFalse(n.IsColVector);
            Assert.IsFalse(n.IsRowVector);
            Assert.IsTrue(n.IsMatrix);

            n.Dispose();
        }

        [Test]
        public void FromColVector()
        {
            double[] d = new double[3] {0, 1, 2};
            Matrix n = Matrix.FromVector(d);

            Assert.AreEqual(n.Rows, 3);
            Assert.AreEqual(n.Cols, 1);
            Assert.AreEqual(d[0], n[0, 0]);
            Assert.AreEqual(d[1], n[1, 0]);
            Assert.AreEqual(d[2], n[2, 0]);
            Assert.AreEqual(d[0], n[0]);
            Assert.AreEqual(d[1], n[1]);
            Assert.AreEqual(d[2], n[2]);
            Assert.IsFalse(n.IsScalar);
            Assert.IsTrue(n.IsVector);
            Assert.IsTrue(n.IsColVector);
            Assert.IsFalse(n.IsRowVector);
            Assert.IsFalse(n.IsMatrix);

            n.Dispose();
        }

        [Test]
        public void FromJaggedArray()
        {
            double[][] d = new double[2][];
            d[0] = new double[3] {0, 1, 2};
            d[1] = new double[3] {3, 4, 5};
            Matrix n = Matrix.FromJaggedArray(d);

            Assert.AreEqual(n.Rows, 2);
            Assert.AreEqual(n.Cols, 3);
            Assert.AreEqual(d[0][0], n[0, 0]);
            Assert.AreEqual(d[0][1], n[0, 1]);
            Assert.AreEqual(d[0][2], n[0, 2]);
            Assert.AreEqual(d[1][0], n[1, 0]);
            Assert.AreEqual(d[1][1], n[1, 1]);
            Assert.AreEqual(d[1][2], n[1, 2]);
            Assert.IsFalse(n.IsScalar);
            Assert.IsFalse(n.IsVector);
            Assert.IsFalse(n.IsColVector);
            Assert.IsFalse(n.IsRowVector);
            Assert.IsTrue(n.IsMatrix);

            n.Dispose();
        }

        [Test]
        public void FromMultiArray()
        {
            double[,] d = new double[2,3] {{0, 1, 2}, {3, 4, 5}};
            Matrix n = Matrix.FromMultiArray(d);

            Assert.AreEqual(n.Rows, 2);
            Assert.AreEqual(n.Cols, 3);
            Assert.AreEqual(d[0, 0], n[0, 0]);
            Assert.AreEqual(d[0, 1], n[0, 1]);
            Assert.AreEqual(d[0, 2], n[0, 2]);
            Assert.AreEqual(d[1, 0], n[1, 0]);
            Assert.AreEqual(d[1, 1], n[1, 1]);
            Assert.AreEqual(d[1, 2], n[1, 2]);
            Assert.IsFalse(n.IsScalar);
            Assert.IsFalse(n.IsVector);
            Assert.IsFalse(n.IsColVector);
            Assert.IsFalse(n.IsRowVector);
            Assert.IsTrue(n.IsMatrix);

            n.Dispose();
        }

        [Test]
        public void FromRowVector()
        {
            double[] d = new double[3] {0, 1, 2};
            Matrix n = Matrix.FromArray(d);

            Assert.AreEqual(n.Rows, 1);
            Assert.AreEqual(n.Cols, 3);
            Assert.AreEqual(d[0], n[0, 0]);
            Assert.AreEqual(d[1], n[0, 1]);
            Assert.AreEqual(d[2], n[0, 2]);
            Assert.AreEqual(d[0], n[0]);
            Assert.AreEqual(d[1], n[1]);
            Assert.AreEqual(d[2], n[2]);
            Assert.IsFalse(n.IsScalar);
            Assert.IsTrue(n.IsVector);
            Assert.IsFalse(n.IsColVector);
            Assert.IsTrue(n.IsRowVector);
            Assert.IsFalse(n.IsMatrix);

            n.Dispose();
        }

        [Test]
        public void FromScalar()
        {
            double scalar = 2.0;
            Matrix n = Matrix.FromScalar(scalar);

            Assert.AreEqual(n.Rows, 1);
            Assert.AreEqual(n.Cols, 1);
            Assert.AreEqual(n[0,0], n[0]);
            Assert.AreEqual(scalar, n[0]);
            Assert.AreEqual(scalar, n.ToScalar());
            Assert.IsTrue(n.IsScalar);
            Assert.IsFalse(n.IsVector);
            Assert.IsFalse(n.IsColVector);
            Assert.IsFalse(n.IsRowVector);
            Assert.IsFalse(n.IsMatrix);

            n.Dispose();
        }

        [Test]
        public void ToArrayRowMajor()
        {
            double[] d = m.ToArray();

            Assert.AreEqual(m[0, 0], d[0]);
            Assert.AreEqual(m[0, 1], d[1]);
            Assert.AreEqual(m[0, 2], d[2]);
            Assert.AreEqual(m[1, 0], d[3]);
            Assert.AreEqual(m[1, 1], d[4]);
            Assert.AreEqual(m[1, 2], d[5]);
        }

        [Test]
        public void ToArrayColMajor()
        {
            double[] d = m.ToArray(false);

            Assert.AreEqual(m[0, 0], d[0]);
            Assert.AreEqual(m[1, 0], d[1]);
            Assert.AreEqual(m[0, 1], d[2]);
            Assert.AreEqual(m[1, 1], d[3]);
            Assert.AreEqual(m[0, 2], d[4]);
            Assert.AreEqual(m[1, 2], d[5]);
        }

        [Test]
        public void ToJaggedArray()
        {
            double[][] d = m.ToJaggedArray();

            Assert.AreEqual(m[0, 0], d[0][0]);
            Assert.AreEqual(m[0, 1], d[0][1]);
            Assert.AreEqual(m[0, 2], d[0][2]);
            Assert.AreEqual(m[1, 0], d[1][0]);
            Assert.AreEqual(m[1, 1], d[1][1]);
            Assert.AreEqual(m[1, 2], d[1][2]);
        }

        [Test]
        public void ToMultiArray()
        {
            double[,] d = m.ToMultiArray();

            Assert.AreEqual(m[0, 0], d[0, 0]);
            Assert.AreEqual(m[0, 1], d[0, 1]);
            Assert.AreEqual(m[0, 2], d[0, 2]);
            Assert.AreEqual(m[1, 0], d[1, 0]);
            Assert.AreEqual(m[1, 1], d[1, 1]);
            Assert.AreEqual(m[1, 2], d[1, 2]);
        }
        #endregion

        #region Exception unit tests
        [Test]
        [ExpectedException(typeof(ObjectDisposedException))]
        public void DisposedAccess()
        {
            Matrix n = new Matrix(2, 2);
            n.Dispose();

            double d = n[0, 0];
        }

        [Test]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void OutOfBoundsAccess1()
        {
            double d = m[3, 0];
        }

        [Test]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void OutOfBoundsAccess2()
        {
            m[0, -1] = 0;
        }

        [Test]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void WrongNumRowsFromArray1()
        {
            double[] d = new double[6] {0, 1, 2, 3, 4, 5};
            Matrix n = Matrix.FromArray(d, 5);
        }

        [Test]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void WrongNumRowsFromArray2()
        {
            double[] d = new double[6] {0, 1, 2, 3, 4, 5};
            Matrix n = Matrix.FromArray(d, 0);
        }

        [Test]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void ZeroSizeCol()
        {
            Matrix n = new Matrix(2, 0);
        }

        [Test]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void ZeroSizeFromArray()
        {
            double[] d = new double[0];
            Matrix n = Matrix.FromArray(d, 1);
        }

        [Test]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void ZeroSizeFromMultiArray()
        {
            double[,] d = new double[0,1];
            Matrix n = Matrix.FromMultiArray(d);
        }

        [Test]
        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        public void ZeroSizeRow()
        {
            Matrix n = new Matrix(0, 2);
        }
        #endregion
    }
}
