using System;
using System.Runtime.InteropServices;

namespace Lydos.Matlab
{
    /// <summary>
    /// Provides access to a Matlab matrix of real doubles.
    /// </summary>
    /// <remarks>
    /// All C# multidimensional arrays are created and accessed in row-column order, i.e. <c>Item[row, col]</c> and <c>Item[row][col]</c>.
    /// Flat C# arrays can be treated as column or row vectors (matrices with one column or one row, respectively),
    /// or as matrices stored in either row-major or column-major order. 
    /// For row major order (the default for C programs), items are accessed "row first" (rows are stored sequentially).
    /// For column major order (the default for Fortran and MATLAB), items are accessed "column first" (columns are stored sequentially).
    /// See the following detailed example for an explanation:
    /// <code>
    /// Matrix:    Col  __0__1__2_
    ///            Row |
    ///              0 |  A  B  C
    ///              1 |  D  E  F
    /// </code>
    /// In row major order, <c>Item[row, col] = Item[row * Cols + col]</c>:
    /// <code>
    /// Item:          A      B      C      D      E      F
    /// Row, Col:     0,0    0,1    0,2    1,0    1,1    1,2
    /// Flat Index:  0*3+0  0*3+1  0*3+1  1*3+0  1*3+1  1*3+2
    ///               = 0    = 1    = 2    = 3    = 4    = 5
    /// </code>
    /// In column major order, <c>Item[row, col] = Item[col * Rows + row]</c>:
    /// <code>
    /// Item:          A      D      B      E      C      F
    /// Row, Col:     0,0    1,0    0,1    1,1    0,2    1,2
    /// Col, Row:     0,0    0,1    1,0    1,1    2,0    2,1
    /// Flat Index:  0*2+0  0*2+1  1*2+0  1*2+1  2*2+0  2*2+1
    ///               = 0    = 1    = 2    = 3    = 4    = 5
    /// </code>
    /// </remarks>
    public class Matrix : IDisposable
    {
        #region Fields
        IntPtr mxPtr;
        int rows, cols;
        #endregion

        #region Constructors
        /// <summary>
        /// Creates an unpopulated matrix of the specified size.
        /// </summary>
        /// <param name="rows">Number of rows.</param>
        /// <param name="cols">Number of columns.</param>
        public Matrix(int rows, int cols)
        {
            CreateEmpty(rows, cols);
        }

        /// <summary>
        /// Creates a matrix by setting it to have the specified mxArray pointer.
        /// </summary>
        /// <param name="matrix">A pointer to an mxArray.</param>
        private Matrix(IntPtr matrix)
        {
            CopyMxPtr(matrix);
        }
        #endregion

        #region Destructor
        /// <summary>
        /// Frees memory taken by the internal mxArray.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Final disposal of object.
        /// </summary>
        /// <param name="disposing">True if called explicitly, false if finalizing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (mxPtr != IntPtr.Zero)
            {
                NativeMethods.mxDestroyArray(mxPtr);
                mxPtr = IntPtr.Zero;
                GC.KeepAlive(this);
            }
        }

        /// <summary>
        /// Finalization.
        /// </summary>
        ~Matrix()
        {
            Dispose(false);
        }
        #endregion

        #region Properties
        /// <summary>
        /// The number of matrix columns.
        /// </summary>
        /// <value>The number of columns in the matrix.</value>
        public int Cols
        {
            get
            {
                return cols;
            }
        }
		
        /// <summary>
        /// Checks whether the matrix is a column vector.
        /// </summary>
        /// <value>True if the matrix is a column vector (Nx1), otherwise false.</value>
        public bool IsColVector 
        {
            get 
            {
                return Cols == 1 && Rows > 1;
            }
        }
		
        /// <summary>
        /// Checks whether the matrix is neither a scalar nor vector.
        /// </summary>
        /// <value>True if the matrix has more than one row and more than one column, otherwise false.</value>
        public bool IsMatrix 
        {
            get 
            {
                return Cols > 1 && Rows > 1;
            }
        }
		
        /// <summary>
        /// Checks whether the matrix is a row vector.
        /// </summary>
        /// <value>True if the matrix is a row vector (1xN), otherwise false.</value>
        public bool IsRowVector 
        {
            get 
            {
                return Rows == 1 && Cols > 1;
            }
        }

        /// <summary>
        /// Checks whether the matrix is a scalar.
        /// </summary>
        /// <value>True if the matrix is a scalar value (a 1x1 matrix), otherwise false.</value>
        public bool IsScalar
        {
            get
            {
                return Cols == 1 && Rows == 1;
            }
        }
		
        /// <summary>
        /// Checks whether the matrix is a vector.
        /// </summary>
        /// <value>True if the matrix is a vector (1xN or Nx1), otherwise false.</value>
        public bool IsVector
        {
            get
            {
                return IsRowVector || IsColVector;
            }
        }

        /// <summary>
        /// The pointer to the mxArray.
        /// </summary>
        /// <value>The internal pointer to the data structure used by Matlab.</value>
        internal IntPtr MxPtr
        {
            get
            {
                return mxPtr;
            }
        }

        /// <summary>
        /// The number of matrix rows.
        /// </summary>
        /// <value>The number of rows in the matrix.</value>
        public int Rows
        {
            get
            {
                return rows;
            }
        }
		
        /// <summary>
        /// Accesses the matrix as a vector.
        /// </summary>
        /// <value>The element at the specified index of the vector.</value>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when <paramref name="index"/> is less than zero or greater than the number of elements.</exception>
        /// <remarks>If this matrix is actually multidimensional, note that MATLAB stores matricies in column major order.</remarks>
        public double this[int index] 
        {
            get 
            {
                if (mxPtr == IntPtr.Zero) 
                {
                    throw new ObjectDisposedException(null, "Cannot access a disposed matrix.");
                }
                if (index < 0 || index >= Rows * Cols) 
                {
                    throw new ArgumentOutOfRangeException();
                }
				
                double val;
				
                unsafe 
                {
                    double* p = (double*)NativeMethods.mxGetPr(mxPtr).ToPointer();
                    val = p[index];
                }
                GC.KeepAlive(this);

                return val;
            }
            set 
            {
                if (index < 0 || index >= Rows * Cols) 
                {
                    throw new ArgumentOutOfRangeException();
                }
				
                unsafe 
                {
                    double* p = (double*)NativeMethods.mxGetPr(mxPtr).ToPointer();
                    p[index] = value;
                }
                GC.KeepAlive(this);
            }

        }

        /// <summary>
        /// Accesses the matrix as a multidimensional array.
        /// </summary>
        /// <value>The element at the specified row and column of the matrix.</value>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when <paramref name="row"/> is less than zero or greater than the number of rows or when <paramref name="col"/> is less than zero or greater than the number of columns.</exception>
        public double this[int row, int col]
        {
            get
            {
                if (mxPtr == IntPtr.Zero)
                {
                    throw new ObjectDisposedException(null, "Cannot access a disposed matrix.");
                }
                if (row < 0 || col < 0 || row >= Rows || col >= Cols)
                {
                    throw new ArgumentOutOfRangeException();
                }

                double val;

                unsafe
                {
                    double* p = (double*)NativeMethods.mxGetPr(mxPtr).ToPointer();
                    val = p[Rows * col + row];
                }
                GC.KeepAlive(this);

                return val;
            }
            set
            {
                if (row < 0 || col < 0 || row >= Rows || col >= Cols)
                {
                    throw new ArgumentOutOfRangeException();
                }

                unsafe
                {
                    double* p = (double*)NativeMethods.mxGetPr(mxPtr).ToPointer();
                    p[Rows * col + row] = value;
                }
                GC.KeepAlive(this);
            }
        }
        #endregion

        #region Methods
        /// <summary>
        /// Copies an array to the mxArray for this matrix.
        /// </summary>
        /// <param name="matrix">Matrix in row major or column major order.</param>
        /// <param name="rowMajor">True if <paramref name="matrix"/> is in row major order, false for column major.</param>
        /// <remarks>This assumes the mxArray and array are the same size.</remarks>
        private void CopyArray(double[] matrix, bool rowMajor)
        {
            // Create transpose to store in Matlab's column major format if necessary
            if (rowMajor)
            {
                double[] matrixTranspose = new double[matrix.Length];

                for (int j = 0, count = 0; j < cols; ++j) 
                {
                    for (int i = 0; i < rows; i++) 
                    {
                        matrixTranspose[count] = matrix[i * cols + j];
                        ++count;
                    }
                }

                Marshal.Copy(matrixTranspose, 0, NativeMethods.mxGetPr(mxPtr), matrixTranspose.Length);
            }
            else
            {
                Marshal.Copy(matrix, 0, NativeMethods.mxGetPr(mxPtr), matrix.Length);
            }
            GC.KeepAlive(this);
        }

        /// <summary>
        /// Copies a jagged array to the mxArray for this matrix.
        /// </summary>
        /// <param name="matrix">The matrix as a jagged array.</param>
        /// <remarks>This assumes the mxArray and jagged array are the same size.</remarks>
        private void CopyJaggedArray(double[][] matrix)
        {
            // Create transpose to store in Matlab's column major format
            double[] matrixTranspose = new double[rows * cols];

            for (int j = 0, count = 0; j < cols; ++j)
            {
                for (int i = 0; i < rows; ++i)
                {
                    matrixTranspose[count] = matrix[i][j];
                    ++count;
                }
            }

            Marshal.Copy(matrixTranspose, 0, NativeMethods.mxGetPr(mxPtr), matrixTranspose.Length);
            GC.KeepAlive(this);
        }

        /// <summary>
        /// Copies a multidimensional array to the mxArray for this matrix.
        /// </summary>
        /// <param name="matrix">The matrix as a multidimensional array.</param>
        /// <remarks>This assumes the mxArray and multidimensional array are the same size.</remarks>
        private void CopyMultiArray(double[,] matrix)
        {
            // Create transpose to store in Matlab's column major format
            double[] matrixTranspose = new double[matrix.Length];

            for (int j = 0, count = 0; j < cols; ++j)
            {
                for (int i = 0; i < rows; ++i)
                {
                    matrixTranspose[count] = matrix[i, j];
                    ++count;
                }
            }

            Marshal.Copy(matrixTranspose, 0, NativeMethods.mxGetPr(mxPtr), matrixTranspose.Length);
            GC.KeepAlive(this);
        }

        /// <summary>
        /// Sets the matrix to have the specified mxArray pointer.
        /// </summary>
        /// <param name="matrix">A pointer to an mxArray.</param>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when <paramref name="matrix"/> is a null pointer.</exception>
        /// <exception cref="MatlabBadTypeException">Thrown when <paramref name="matrix"/> is not a pointer to a real, double Matlab matrix.</exception>
        private void CopyMxPtr(IntPtr matrix)
        {
            if (matrix == IntPtr.Zero)
            {
                throw new ArgumentOutOfRangeException("matrix", "Cannot set mxArray pointer to null.");
            }
            if (NativeMethods.mxGetClassID(matrix) != NativeMethods.mxClassID.Double)
            {
                throw new MatlabBadTypeException("Matrix class can only represent double matrices");
            }
            if (NativeMethods.mxIsComplex(matrix))
            {
                throw new MatlabBadTypeException("Matrix class can only represent real matrices");
            }

            mxPtr = matrix;

            this.rows = NativeMethods.mxGetM(mxPtr);
            this.cols = NativeMethods.mxGetN(mxPtr);
            if (this.rows < 1 || this.cols < 1)
            {
                throw new MatlabBadTypeException("Matrix must be at least 1x1");
            }
            GC.KeepAlive(this);
        }

        /// <summary>
        /// Creates an empty matrix filled with zeros.
        /// </summary>
        /// <param name="rows">The number of rows.</param>
        /// <param name="cols">The number of columns.</param>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when <paramref name="rows"/> or <paramref name="cols"/> is less than one.</exception>
        /// <exception cref="MatlabOutOfMemoryException">Thrown when Matlab runs out of memory.</exception>
        private void CreateEmpty(int rows, int cols)
        {
            if (rows < 1)
            {
                throw new ArgumentOutOfRangeException("rows", "Matrix must have a positive number of rows and columns.");
            }
            if (cols < 1)
            {
                throw new ArgumentOutOfRangeException("cols", "Matrix must have a positive number of columns.");
            }

            mxPtr = NativeMethods.mxCreateDoubleMatrix(rows, cols, NativeMethods.mxComplexity.Real);

            if (mxPtr == IntPtr.Zero)
            {
                throw new MatlabOutOfMemoryException();
            }

            this.rows = rows;
            this.cols = cols;

            GC.KeepAlive(this);
        }

        /// <summary>
        /// Creates a row major array copy of the matrix.
        /// </summary>
        /// <remarks>This is the same as <c>ToArray(true)</c>.</remarks>
        public double[] ToArray()
        {
            return ToArray(true);
        }

        /// <summary>
        /// Creates an array copy of the matrix.
        /// </summary>
        /// <param name="rowMajor">True to return the matrix in row major order, otherwise false.</param>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        /// <remarks>Getting a row major matrix requires taking a transpose, and so it is slightly more expensive.</remarks>
        public double[] ToArray(bool rowMajor)
        {
            if (mxPtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException(null, "Cannot access disposed matrix.");
            }

            double[] matrix = new double[rows * cols];

            Marshal.Copy(NativeMethods.mxGetPr(mxPtr), matrix, 0, rows * cols);
            GC.KeepAlive(this);

            // Matlab uses column major, convert to row major by taking transpose
            if (rowMajor)
            {
                double[] matrixTranspose = new double[matrix.Length];

                for (int i = 0, count = 0; i < rows; ++i)
                {
                    for (int j = 0; j < cols; ++j)
                    {
                        matrixTranspose[count] = matrix[j * rows + i];
                        ++count;
                    }
                }

                matrix = matrixTranspose;
            }

            return matrix;
        }

        /// <summary>
        /// Creates a jagged array copy of the matrix.
        /// </summary>
        /// <returns>A jagged array corresponding to the matrix.</returns>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        public double[][] ToJaggedArray()
        {
            if (mxPtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException(null, "Cannot access disposed matrix.");
            }

            double[][] matrix = new double[rows][];
            for (int i = 0; i < rows; ++i)
            {
                matrix[i] = new double[cols];
            }
            double[] matrixTranspose = ToArray(false);

            // Convert to row major and copy to jagged array
            for (int j = 0, count = 0; j < cols; ++j)
            {
                for (int i = 0; i < rows; ++i)
                {
                    matrix[i][j] = matrixTranspose[count];
                    ++count;
                }
            }

            return matrix;
        }

        /// <summary>
        /// Creates a multidimensional array copy of the matrix.
        /// </summary>
        /// <returns>A multidimensional array corresponding to the matrix.</returns>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        public double[,] ToMultiArray()
        {
            if (mxPtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException(null, "Cannot access disposed matrix.");
            }

            double[,] matrix = new double[rows, cols];
            double[] matrixTranspose = ToArray(false);

            // Convert to row major and copy to multidimensional array
            for (int j = 0, count = 0; j < cols; ++j)
            {
                for (int i = 0; i < rows; ++i)
                {
                    matrix[i, j] = matrixTranspose[count];
                    ++count;
                }
            }

            return matrix;
        }

        /// <summary>
        /// Accesses the matrix as a single scalar value.
        /// </summary>
        /// <returns>The first element of the matrix.</returns>
        public double ToScalar() 
        {
            return this[0];
        }

        /// <summary>
        /// Creates and returns a string representation of the matrix.
        /// </summary>
        /// <returns>A string representation of the matrix in a Matlab readable format.</returns>
        /// <exception cref="ObjectDisposedException">Thrown when this has already been disposed.</exception>
        public override string ToString()
        {
            if (mxPtr == IntPtr.Zero)
            {
                throw new ObjectDisposedException(null, "Cannot access disposed matrix.");
            }

            System.Text.StringBuilder s = new System.Text.StringBuilder();
            int rowsMinusOne = rows - 1;
            int colsMinusOne = cols - 1;

            s.Append("[");
            for (int i = 0; i < rowsMinusOne; ++i)
            {
                for (int j = 0; j < colsMinusOne; ++j)
                {
                    s.Append(this[i, j]);
                    s.Append(" ");
                }
                s.Append(this[i, colsMinusOne]);
                s.Append("; ");
            }
            for (int j = 0; j < colsMinusOne; ++j)
            {
                s.Append(this[rowsMinusOne, j]);
                s.Append(" ");
            }
            s.Append(this[rowsMinusOne, colsMinusOne]);
            s.Append("]");

            return s.ToString();
        }
		
        /// <summary>
        /// Creates a column major array copy of the matrix.
        /// </summary>
        /// <remarks>This is the same as <c>ToArray(false)</c>.</remarks>
        public double[] ToVector() 
        {
            return ToArray(false);
        }
        #endregion

        #region Static methods
        /// <summary>
        /// Creates a column vector from an array.
        /// </summary>
        /// <param name="matrix">A column vector.</param>
        public static Matrix FromVector(double[] matrix)
        {
            return FromArray(matrix, 1, false);
        }

        /// <summary>
        /// Creates a row vector from an array.
        /// </summary>
        /// <param name="matrix">A row vector.</param>
        public static Matrix FromArray(double[] matrix)
        {
            return FromArray(matrix, matrix.Length, true);
        }

        /// <summary>
        /// Creates a matrix from a row major array.
        /// </summary>
        /// <param name="matrix">Matrix in row major order.</param>
        /// <param name="cols">The number of columns.</param>
        public static Matrix FromArray(double[] matrix, int cols)
        {
            return FromArray(matrix, cols, true);
        }

        /// <summary>
        /// Creates a matrix from an array.
        /// </summary>
        /// <param name="matrix">Matrix in row major or column major order.</param>
        /// <param name="cols">The number of columns.</param>
        /// <param name="rowMajor">True if <paramref name="matrix"/> is in row major order, false for column major.</param>
        /// <returns>A matrix corresponding to the specified array.</returns>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="matrix"/> is null.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when <paramref name="cols"/> is less than one or matrix does not have a usable number of elements.</exception>
        public static Matrix FromArray(double[] matrix, int cols, bool rowMajor)
        {
            if (matrix == null)
            {
                throw new ArgumentNullException("matrix");
            }
            if (cols < 1)
            {
                throw new ArgumentOutOfRangeException("cols", "Must have a positive number of columns.");
            }
            if (matrix.Length < 1)
            {
                throw new ArgumentOutOfRangeException("matrix", "Array must have some elements.");
            }

            int rows = matrix.Length / cols;

            if (rows * cols != matrix.Length)
            {
                throw new ArgumentOutOfRangeException("matrix", "Array's size does not match number of columns specified.");
            }

            Matrix m = new Matrix(rows, cols);
            m.CopyArray(matrix, rowMajor);
            return m;
        }

        /// <summary>
        /// Creates a matrix from a jagged array.
        /// </summary>
        /// <param name="matrix">The matrix as a jagged array.</param>
        /// <returns>A matrix corresponding to the specified jagged array.</returns>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="matrix"/> or any of its subarrays are null.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when <paramref name="matrix"/> is empty.</exception>
        public static Matrix FromJaggedArray(double[][] matrix)
        {
            if (matrix == null)
            {
                throw new ArgumentNullException("matrix");
            }
            if (matrix.Length < 1)
            {
                throw new ArgumentOutOfRangeException("matrix", "Matrix must have some elements.");
            }

            Matrix m = new Matrix(matrix.Length, matrix[0].Length);
            m.CopyJaggedArray(matrix);
            return m;
        }

        /// <summary>
        /// Creates a matrix from a multidimensional array.
        /// </summary>
        /// <param name="matrix">The matrix as a multidimensional array.</param>
        /// <returns>A matrix corresponding to the specified multidimensional array.</returns>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="matrix"/> is null.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when <paramref name="matrix"/> is empty.</exception>
        public static Matrix FromMultiArray(double[,] matrix)
        {
            if (matrix == null)
            {
                throw new ArgumentNullException("matrix");
            }
            if (matrix.Length < 1)
            {
                throw new ArgumentOutOfRangeException("matrix", "Matrix must have some elements.");
            }

            Matrix m = new Matrix(matrix.GetLength(0), matrix.GetLength(1));
            m.CopyMultiArray(matrix);
            return m;
        }

        /// <summary>
        /// Creates a matrix from the specified mxArray pointer.
        /// </summary>
        /// <param name="matrix">A pointer to an mxArray.</param>
        /// <returns>A matrix corresponding to the specified mxArray pointer.</returns>
        internal static Matrix FromMxPtr(IntPtr matrix)
        {
            Matrix m = new Matrix(matrix);
            return m;
        }

        /// <summary>
        /// Creates a 1x1 matrix from a scalar.
        /// </summary>
        /// <param name="scalar">The 1x1 matrix as a single value.</param>
        /// <returns>A matrix corresponding to the specified scalar.</returns>
        public static Matrix FromScalar(double scalar)
        {
            Matrix m = new Matrix(1, 1);
            m[0] = scalar;
            return m;
        }
        #endregion
    }
}
